@@ -42,7 +42,7 @@ struct uwsgi_gateway *register_gateway(char *name, void (*loop) (int, void *), v
 }
 
 static void gateway_brutal_end() {
-	_exit(UWSGI_END_CODE);
+        _exit(UWSGI_END_CODE);
 }
 
 void gateway_respawn(int id) {
@@ -82,7 +82,7 @@ void gateway_respawn(int id) {
 
 		ug->loop(id, ug->data);
 		// never here !!! (i hope)
-		_exit(1);
+		exit(1);
 	}
 
 	ug->pid = gw_pid;
