@@ -91,7 +91,6 @@ retry:
 	}
 
 #ifdef EOWNERDEAD
-
 #ifndef PTHREAD_MUTEX_ROBUST
 #define PTHREAD_MUTEX_ROBUST PTHREAD_MUTEX_ROBUST_NP
 #endif
@@ -402,6 +401,60 @@ void uwsgi_rwunlock_fast(struct uwsgi_lock_item *uli) {
 	uwsgi_unlock_fast(uli);
 }
 
+#elif defined(UWSGI_LOCK_USE_WINDOWS_MUTEX)
+
+#define UWSGI_LOCK_ENGINE_NAME "windows mutexes"
+#define UWSGI_LOCK_SIZE         sizeof(HANDLE)
+#define UWSGI_RWLOCK_SIZE       sizeof(HANDLE)
+
+
+struct uwsgi_lock_item *uwsgi_lock_fast_init(char *id) {
+
+        struct uwsgi_lock_item *uli = uwsgi_register_lock(id, 0);
+	struct _SECURITY_ATTRIBUTES sa;
+	memset(&sa, 0, sizeof(struct _SECURITY_ATTRIBUTES));
+	sa.bInheritHandle = 1;
+	uli->lock_ptr = CreateMutex(&sa, FALSE, NULL);
+        return uli;
+}
+
+void uwsgi_lock_fast(struct uwsgi_lock_item *uli) {
+	WaitForSingleObject(uli->lock_ptr, INFINITE);
+        uli->pid = uwsgi.mypid;
+}
+
+void uwsgi_unlock_fast(struct uwsgi_lock_item *uli) {
+	ReleaseMutex(uli->lock_ptr);
+        uli->pid = 0;
+}
+
+pid_t uwsgi_lock_fast_check(struct uwsgi_lock_item *uli) {
+	if (WaitForSingleObject(uli->lock_ptr, 0) == WAIT_TIMEOUT) {
+		return 0;
+	}
+        return uli->pid;
+}
+
+struct uwsgi_lock_item *uwsgi_rwlock_fast_init(char *id) {
+	return uwsgi_lock_fast_init(id);
+}
+
+void uwsgi_rlock_fast(struct uwsgi_lock_item *uli) {
+	uwsgi_lock_fast(uli);
+}
+void uwsgi_wlock_fast(struct uwsgi_lock_item *uli) {
+	uwsgi_lock_fast(uli);
+}
+
+pid_t uwsgi_rwlock_fast_check(struct uwsgi_lock_item *uli) {
+	return uwsgi_lock_fast_check(uli);
+}
+
+void uwsgi_rwunlock_fast(struct uwsgi_lock_item *uli) {
+	uwsgi_unlock_fast(uli);
+}
+
+
 #else
 
 #define uwsgi_lock_fast_init uwsgi_lock_ipcsem_init
@@ -480,7 +533,7 @@ void uwsgi_lock_ipcsem(struct uwsgi_lock_item *uli) {
 
 retry:
 	if (semop(semid, &sb, 1)) {
-		if (errno == EINTR) goto retry;
+		if (errno == EINTR) goto retry; 
 		uwsgi_error("semop()");
 	}
 }
@@ -497,7 +550,7 @@ void uwsgi_unlock_ipcsem(struct uwsgi_lock_item *uli) {
 
 retry:
 	if (semop(semid, &sb, 1)) {
-		if (errno == EINTR) goto retry;
+		if (errno == EINTR) goto retry; 
 		uwsgi_error("semop()");
 	}
 
@@ -562,6 +615,10 @@ pid_t uwsgi_rwlock_ipcsem_check(struct uwsgi_lock_item *uli) {
 
 void uwsgi_setup_locking() {
 
+	int i;
+
+	if (uwsgi.locking_setup) return;
+
 	// use the fastest available locking
 	if (uwsgi.lock_engine) {
 		if (!strcmp(uwsgi.lock_engine, "ipcsem")) {
@@ -600,9 +657,8 @@ void uwsgi_setup_locking() {
 	uwsgi.lock_size = UWSGI_LOCK_SIZE;
 	uwsgi.rwlock_size = UWSGI_RWLOCK_SIZE;
 
-	// application generic lock
-	int i;
 ready:
+	// application generic lock
 	uwsgi.user_lock = uwsgi_malloc(sizeof(void *) * (uwsgi.locks + 1));
 	for (i = 0; i < uwsgi.locks + 1; i++) {
 		uwsgi.user_lock[i] = uwsgi_lock_init(uwsgi_concat2("user ", uwsgi_num2str(i)));
@@ -635,11 +691,20 @@ ready:
 
 	if (uwsgi.use_thunder_lock) {
 		// process shared thunder lock
-		uwsgi.the_thunder_lock = uwsgi_lock_init("thunder");
+		uwsgi.the_thunder_lock = uwsgi_lock_init("thunder");	
 	}
 
 	uwsgi.rpc_table_lock = uwsgi_lock_init("rpc");
 
+#ifdef UWSGI_SSL
+	// register locking for legions
+	struct uwsgi_legion *ul = uwsgi.legions;
+	while(ul) {
+		ul->lock = uwsgi_lock_init(uwsgi_concat2("legion_", ul->legion));
+		ul = ul->next;
+	}
+#endif
+	uwsgi.locking_setup = 1;
 }
 
 
